/*
 * **********************************
 *                                   *
 * HLC Games - Software Entertainment *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC BrickCoverage - (1.6.2010) ************************
 *
 * (c) Copyright 2010 - Lukas Hermanns
 *
 * Basic brick file
 *
 */

#include "bcBasicBrick.hpp"


namespace bc
{


/*
 * Global members
 */

BasicBrick* CurDragedBrick = 0;
std::list<BasicBrick*> BrickObjectList;
s32 CurDragedRotation = 0;

scene::Mesh* BasicBrickObjectList[BRICK_COUNT]  = { 0 };
scene::Mesh* BasicBrickMenuList[BRICK_COUNT]    = { 0 };
dim::aabbox3df BasicBrickBoxList[BRICK_COUNT];

const dim::vector3df BasicBrick::BRICK_DIMENSION = dim::vector3df(0.5f, 0.2f, 0.5f);

video::color BasicColorTypes[COLOR_COUNT] = {
    video::color(255),
    video::color(50),
    video::color(255, 0, 0),
    video::color(0, 255, 0),
    video::color(0, 0, 255),
    video::color(255, 255, 0),
    video::color(255, 200, 0),
    video::color(128),
    video::color(190, 140, 32),
    video::color(255, 255, 255, 128),
};


/*
 * BasicBrick class
 */

BasicBrick::BasicBrick(
    scene::Mesh* BrickMesh, const EBasicBricks BasicType, const dim::vector3di &Size, const dim::aabbox3df &BoundBox)
    : Mesh_(0), PickObj_(0), BasicType_(BasicType), Size_(Size), Rotation_(CurDragedRotation),
    BoundBox_(BoundBox), ColorType_(COLOR_WHITE), Active_(false)
{
    // Create the mesh isntance
    Mesh_ = spSmngr->createMesh();
    Mesh_->setMeshReference(BrickMesh);
    Mesh_->setColorMaterial(false);
    Mesh_->setRotation(dim::vector3df(0, (f32)CurDragedRotation, 0));
    Mesh_->setUserData(this);
    
    UpdateBricks();
    
    // Create the picking object
    PickObj_ = spCmngr->addPickingObject(Mesh_, scene::PICKMODE_BOX, 0.5f, BoundBox);
}
BasicBrick::~BasicBrick()
{
    spCmngr->removePickingObject(PickObj_);
    
    spSmngr->deleteNode(Mesh_);
}

void BasicBrick::Drag()
{
    CurDragedBrick = this;
}
void BasicBrick::Drop()
{
    CurDragedBrick = 0;
}

void BasicBrick::SetPosition(const dim::vector3df &Position)
{
    dim::vector3df Pos;
    
    Pos.X = (f32)((s32)(Position.X*2 + 0.5))*0.5 - 0.25;
    Pos.Y = Position.Y * BRICK_DIMENSION.Y;//(f32)((s32)(Position.Y*5 + 0.5))/5 * Size_.Y;
    Pos.Z = (f32)((s32)(Position.Z*2 + 0.5))*0.5 - 0.25;
    
    Mesh_->setPosition(Pos);
    
    Position_ = (Pos / BRICK_DIMENSION + 0.5).cast<s32>();
}
void BasicBrick::SetPosition(const dim::vector3di &Position)
{
    Position_ = Position;
    Mesh_->setPosition(
        dim::vector3df((f32)Position_.X - 0.5, (f32)Position_.Y, (f32)Position_.Z - 0.5) * BRICK_DIMENSION
    );
}

void BasicBrick::SetRotation(s32 Angle)
{
    Rotation_ = Angle;
    Mesh_->setRotation(dim::vector3df(0, (f32)Rotation_, 0));
}

void BasicBrick::TurnCCW()
{
    Rotation_ -= 90;
    
    if (Rotation_ == -90)
        Rotation_ = 270;
    
    Mesh_->setRotation(dim::vector3df(0, (f32)Rotation_, 0));
    CurDragedRotation = Rotation_;
}
void BasicBrick::TurnCW()
{
    Rotation_ += 90;
    
    if (Rotation_ == 360)
        Rotation_ = 0;
    
    Mesh_->setRotation(dim::vector3df(0, (f32)Rotation_, 0));
    CurDragedRotation = Rotation_;
}

void BasicBrick::Move(s32 Horz, s32 Vert)
{
    Position_.X += Horz;
    Position_.Z += Vert;
    Mesh_->translate(dim::vector3df(Horz, 0, Vert) * BRICK_DIMENSION);
}

void BasicBrick::SetColorType(const EBrickColors Type)
{
    if (Type >= COLOR_WHITE && Type < COLOR_EXTENDED)
    {
        ColorType_  = Type;
        Color_      = BasicColorTypes[Type];
        
        Mesh_->setDiffuseColor(Color_);
        Mesh_->setAmbientColor(Color_ / 4);
    }
}
void BasicBrick::SetColor(const video::color &Color)
{
    ColorType_  = COLOR_EXTENDED;
    Color_      = Color;
    
    Mesh_->setDiffuseColor(Color);
    Mesh_->setAmbientColor(Color / 4);
}

void BasicBrick::SetBasicBrickType(const EBasicBricks Type)
{
    scene::Mesh* BrickObj   = 0;
    dim::vector3di Size;
    
    if (Type >= BRICK_1_1_1 && Type <= BRICK_4_3_2)
    {
        BrickObj = BasicBrickObjectList[Type];
        BoundBox_ = BasicBrickBoxList[Type];
    }
    
    switch (Type)
    {
        case BRICK_1_1_1: Size = dim::vector3di(1, 1, 1); break;
        case BRICK_1_3_1: Size = dim::vector3di(1, 3, 1); break;
        case BRICK_1_1_2: Size = dim::vector3di(1, 1, 2); break;
        case BRICK_1_3_2: Size = dim::vector3di(1, 3, 2); break;
        case BRICK_2_1_2: Size = dim::vector3di(2, 1, 2); break;
        case BRICK_2_3_2: Size = dim::vector3di(2, 3, 2); break;
        case BRICK_3_1_2: Size = dim::vector3di(3, 1, 2); break;
        case BRICK_3_3_2: Size = dim::vector3di(3, 3, 2); break;
        case BRICK_4_1_2: Size = dim::vector3di(4, 1, 2); break;
        case BRICK_4_3_2: Size = dim::vector3di(4, 3, 2); break;
        default:
            PrintError("Invalid basic brick type [" + io::stringc(Type) + "]");
            return;
    }
    
    if (BrickObj)
    {
        Mesh_->setMeshReference(BrickObj);
        
        PickObj_->BoundingBox   = BoundBox_;
        BasicType_              = Type;
        Size_                   = Size;
    }
}

void BasicBrick::Update()
{
    if (Active_ && !CurDragedBrick)
    {
        DrawBoundBox();
        
        if (spControl->keyHit(io::KEY_END))
            TurnCCW();
        if (spControl->keyHit(io::KEY_HOME))
            TurnCW();
        
        const s32 MoveSpeed = (spControl->keyDown(io::KEY_CONTROL) ? 3 : 1);
        
        if (spControl->keyHit(io::KEY_LEFT))
            Move(-MoveSpeed, 0);
        if (spControl->keyHit(io::KEY_RIGHT))
            Move(MoveSpeed, 0);
        if (spControl->keyHit(io::KEY_UP))
            Move(0, MoveSpeed);
        if (spControl->keyHit(io::KEY_DOWN))
            Move(0, -MoveSpeed);
    }
}

void BasicBrick::DrawBoundBox(const video::color &Color)
{
    const dim::matrix4f Mat(Mesh_->getGlobalLocation());
    
    static const f32 Dist = 0.025f;
    const dim::aabbox3df BoundBox(BoundBox_.Min - Dist, BoundBox_.Max + Dist);
    
    const dim::vector3df LeftBottomFront (Mat * dim::vector3df(BoundBox.Min.X, BoundBox.Min.Y, BoundBox.Min.Z));
    const dim::vector3df LeftBottomBack  (Mat * dim::vector3df(BoundBox.Min.X, BoundBox.Min.Y, BoundBox.Max.Z));
    const dim::vector3df RightBottomFront(Mat * dim::vector3df(BoundBox.Max.X, BoundBox.Min.Y, BoundBox.Min.Z));
    const dim::vector3df RightBottomBack (Mat * dim::vector3df(BoundBox.Max.X, BoundBox.Min.Y, BoundBox.Max.Z));
    const dim::vector3df LeftTopFront    (Mat * dim::vector3df(BoundBox.Min.X, BoundBox.Max.Y, BoundBox.Min.Z));
    const dim::vector3df LeftTopBack     (Mat * dim::vector3df(BoundBox.Min.X, BoundBox.Max.Y, BoundBox.Max.Z));
    const dim::vector3df RightTopFront   (Mat * dim::vector3df(BoundBox.Max.X, BoundBox.Max.Y, BoundBox.Min.Z));
    const dim::vector3df RightTopBack    (Mat * dim::vector3df(BoundBox.Max.X, BoundBox.Max.Y, BoundBox.Max.Z));
    
    spDriver->draw3DLine(LeftBottomFront, LeftBottomBack, Color);
    spDriver->draw3DLine(RightBottomFront, RightBottomBack, Color);
    spDriver->draw3DLine(LeftTopFront, LeftTopBack, Color);
    spDriver->draw3DLine(RightTopFront, RightTopBack, Color);
    
    spDriver->draw3DLine(LeftBottomFront, RightBottomFront, Color);
    spDriver->draw3DLine(LeftBottomBack, RightBottomBack, Color);
    spDriver->draw3DLine(LeftTopFront, RightTopFront, Color);
    spDriver->draw3DLine(LeftTopBack, RightTopBack, Color);
    
    spDriver->draw3DLine(LeftBottomFront, LeftTopFront, Color);
    spDriver->draw3DLine(RightBottomFront, RightTopFront, Color);
    spDriver->draw3DLine(LeftBottomBack, LeftTopBack, Color);
    spDriver->draw3DLine(RightBottomBack, RightTopBack, Color);
}


void BasicBrick::LoadBasicBricks()
{
    // Load all basic brick objects
    BasicBrickObjectList[BRICK_1_1_1] = spSmngr->loadMesh("../models/Brick(1,1,1).obj");
    BasicBrickObjectList[BRICK_1_3_1] = spSmngr->loadMesh("../models/Brick(1,3,1).obj");
    BasicBrickObjectList[BRICK_1_1_2] = spSmngr->loadMesh("../models/Brick(1,1,2).obj");
    BasicBrickObjectList[BRICK_1_3_2] = spSmngr->loadMesh("../models/Brick(1,3,2).obj");
    BasicBrickObjectList[BRICK_2_1_2] = spSmngr->loadMesh("../models/Brick(2,1,2).obj");
    BasicBrickObjectList[BRICK_2_3_2] = spSmngr->loadMesh("../models/Brick(2,3,2).obj");
    BasicBrickObjectList[BRICK_3_1_2] = spSmngr->loadMesh("../models/Brick(3,1,2).obj");
    BasicBrickObjectList[BRICK_3_3_2] = spSmngr->loadMesh("../models/Brick(3,3,2).obj");
    BasicBrickObjectList[BRICK_4_1_2] = spSmngr->loadMesh("../models/Brick(4,1,2).obj");
    BasicBrickObjectList[BRICK_4_3_2] = spSmngr->loadMesh("../models/Brick(4,3,2).obj");
    
    // Default confogiration
    for (u32 i = BRICK_1_1_1; i <= BRICK_4_3_2; ++i)
    {
        BasicBrickObjectList[i]->setVisible(false);
        BasicBrickBoxList[i] = BasicBrickObjectList[i]->getMeshBoundingBox();
        
        BasicBrickMenuList[i] = spSmngr->copyNode(BasicBrickObjectList[i]);
        BasicBrickMenuList[i]->centerOrigin();
    }
}

void BasicBrick::UpdateBricks()
{
    // Update each brick object
    spDriver->setLineSize(3);
    spDriver->beginDrawing3D();
    
    for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end(); ++it)
        (*it)->Update();
    
    spDriver->endDrawing3D();
    spDriver->setLineSize(1);
    
    if (!IsPickedMenu)
    {
        // Make intersection tests
        std::list<scene::SPickingContact> PickList = spCmngr->pickIntersection(
            Cam->getPickingLine(spControl->getCursorPosition(), 1000)
        );
        
        BasicBrick* CurBrick = 0;
        
        if (CurDragedBrick)
        {
            for (std::list<scene::SPickingContact>::iterator it = PickList.begin(); it != PickList.end(); ++it)
            {
                if (it->Object == CurDragedBrick->GetMesh())
                    continue;
                
                CurBrick = (BasicBrick*)it->Object->getUserData();
                
                if (CurBrick)
                    it->Point.Y = CurBrick->GetPosition().Y + CurBrick->GetSize().Y;
                
                CurDragedBrick->SetPosition(it->Point);
                
                return;
            }
        }
        else
        {
            for (std::list<scene::SPickingContact>::iterator it = PickList.begin(); it != PickList.end(); ++it)
            {
                CurBrick = (BasicBrick*)it->Object->getUserData();
                
                if (CurBrick)
                {
                    spDriver->setLineSize(3);
                    spDriver->beginDrawing3D();
                    
                    CurBrick->DrawBoundBox(video::color(255, 255, 0));
                    
                    spDriver->endDrawing3D();
                    spDriver->setLineSize(1);
                    
                    if (spControl->mouseHit(io::MOUSE_LEFT))
                        CurBrick->SetActive(!CurBrick->GetActive());
                    break;
                }
            }
        }
    }
    
    IsPickedMenu = false;
}

BasicBrick* BasicBrick::Create(const EBasicBricks Type, bool isLoaded)
{
    scene::Mesh* BrickObj   = 0;
    BasicBrick* NewBrick    = 0;
    dim::vector3di Size;
    dim::aabbox3df BoundBox;
    
    if (Type >= BRICK_1_1_1 && Type <= BRICK_4_3_2)
    {
        BrickObj = BasicBrickObjectList[Type];
        BoundBox = BasicBrickBoxList[Type];
    }
    
    switch (Type)
    {
        case BRICK_1_1_1: Size = dim::vector3di(1, 1, 1); break;
        case BRICK_1_3_1: Size = dim::vector3di(1, 3, 1); break;
        case BRICK_1_1_2: Size = dim::vector3di(1, 1, 2); break;
        case BRICK_1_3_2: Size = dim::vector3di(1, 3, 2); break;
        case BRICK_2_1_2: Size = dim::vector3di(2, 1, 2); break;
        case BRICK_2_3_2: Size = dim::vector3di(2, 3, 2); break;
        case BRICK_3_1_2: Size = dim::vector3di(3, 1, 2); break;
        case BRICK_3_3_2: Size = dim::vector3di(3, 3, 2); break;
        case BRICK_4_1_2: Size = dim::vector3di(4, 1, 2); break;
        case BRICK_4_3_2: Size = dim::vector3di(4, 3, 2); break;
        default:
            PrintError("Invalid basic brick type [" + io::stringc(Type) + "]");
            return 0;
    }
    
    if (BrickObj)
    {
        NewBrick = new BasicBrick(BrickObj, Type, Size, BoundBox);
        BrickObjectList.push_back(NewBrick);
        
        if (!isLoaded)
            NewBrick->Drag();
    }
    
    return NewBrick;
}

void BasicBrick::DrawMenuCurrentBrick()
{
    static video::Texture* MenuTexCur;
    
    if (!MenuTexCur)
    {
        MenuTexCur = spDriver->createTexture(128);
        MenuTexCur->makeRenderTarget();
    }
    
    spDriver->setRenderTarget(MenuTexCur);
    spDriver->clearBuffers();
    
    dim::matrix4f Mat;
    spDriver->setViewMatrix(Mat);
    spDriver->setWorldMatrix(Mat);
    
    //Lit->update();
    
    Mat.setPerspectiveRH(scene::DEF_PERSPECTIVE_FOV, 1.0f, 0.1f, 100.0f);
    spDriver->setProjectionMatrix(Mat);
    spDriver->setViewport(0, 0, MenuTexCur->getSize().Width, MenuTexCur->getSize().Height);
    
    BasicBrickMenuList[CurBasicBrickType]->setPosition(dim::vector3df(0, 0, 2.5));
    
    Mat.reset();
    Mat.rotateX(-45);
    Mat.rotateY(CurDragedRotation);
    BasicBrickMenuList[CurBasicBrickType]->setRotationMatrix(Mat);
    
    spDriver->enableVertexBuffer();
    BasicBrickMenuList[CurBasicBrickType]->render();
    spDriver->disableVertexBuffer();
    
    spDriver->setRenderTarget(0);
    
    spDriver->beginDrawing2D(true);
    spDriver->draw2DImage(
        MenuTexCur,
        dim::point2di(
            ScrSize.Width/2 - MenuTexCur->getSize().Width/2,
            ScrSize.Height - MenuTexCur->getSize().Height - 15
        )
    );
    spDriver->endDrawing2D();
}

bool BasicBrick::DrawMenuColorPalette(EBrickColors &ChoosenColor)
{
    spDriver->beginDrawing2D();
    
    bool Result = false;
    
    dim::rect2di Rect, FinalRect;
    
    Rect.Left   = 5;
    Rect.Right  = 55;
    Rect.Top    = 5;
    Rect.Bottom = 25;
    
    for (s32 i = 0; i < COLOR_EXTENDED; ++i, Rect.Top += 20, Rect.Bottom += 20)
    {
        FinalRect = Rect;
        
        if (Rect.isPointCollided(spControl->getCursorPosition()))
        {
            FinalRect.Left      += 3;
            FinalRect.Top       += 3;
            FinalRect.Right     -= 3;
            FinalRect.Bottom    -= 3;
            
            if (spControl->mouseHit(io::MOUSE_LEFT))
            {
                ChoosenColor = (EBrickColors)i;
                Result = true;
            }
            
            IsPickedMenu = true;
        }
        
        if (BasicColorTypes[i] == BackgroundColor)
        {
            spDriver->draw2DRectangle(FinalRect, 0);
            spDriver->draw2DRectangle(FinalRect + dim::rect2di(1, 1, -1, -1), BasicColorTypes[i]);
        }
        else
            spDriver->draw2DRectangle(FinalRect, BasicColorTypes[i]);
    }
    
    spDriver->endDrawing2D();
    
    return Result;
}


/*
 * ======= Protected: =======
 */

void BasicBrick::InitMesh()
{
    
}


} // /namespace bc







// ================================================================================
