/*
 * **********************************
 *                                   *
 * HLC Games - Software Entertainment *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC BrickCoverage - (1.6.2010) ************************
 *
 * (c) Copyright 2010 - Lukas Hermanns
 *
 * BrickCoverage game file
 *
 */

#include "bcGame.hpp"


int main(int argc, char* argv[])
{
    bc::ExamineArguments(argc, argv);
    
    bc::InitGame();
    
    bc::LoadResources();
    
    bc::RenderGame();
    
    bc::ClearGame();
    
    return 0;
}


namespace bc
{


void ExamineArguments(s32 ArgumentCount, c8* Arguments[])
{
    io::stringc Arg;
    
    for (s32 i = 0; i < ArgumentCount; ++i)
    {
        Arg = io::stringc(Arguments[i]);
        
        //if (Arg == "")
        //    ;
        
        // to be continued ...
    }
}

void InitGame()
{
    // Create the device context
    spDevice    = createGraphicsDevice(RendererType, ScrSize, 32, "BrickCoverage", IsFullScreen);
    
    spDriver    = spDevice->getVideoDriver();
    spSmngr     = spDevice->getSceneManager();
    spCmngr     = spDevice->getCollisionDetector();
    spControl   = spDevice->getInputControl();
    
    // Default configuration
    spControl->setCursorVisible(true);
    spDriver->setClearColor(BackgroundColor);
    
    // Create the camera & light
    Cam = spSmngr->createCamera();
    Cam->setPosition(dim::vector3df(0, 20, -20));
    Cam->lookAt(0.0f);
    
    Lit = spSmngr->createLight();
    Lit->setRotation(dim::vector3df(70, 25, 0));
    
    spSmngr->setLighting();
    
    math::SeedRandom();
    
    // Print information header
    io::printMessage("########################################");
    io::printMessage("#                                      #");
    io::printMessage("#     Brick Coverage - v.1.0 alpha     #");
    io::printMessage("#                                      #");
    io::printMessage("#  HLC Games - Software Entertainment  #");
    io::printMessage("# Copyright (c) 2010 by Lukas Hermanns #");
    io::printMessage("#                                      #");
    io::printMessage("########################################");
    io::printMessage("");
}

void LoadResources()
{
    // Load all basic bricks
    BasicBrick::LoadBasicBricks();
    
    // Load all textures
    spDriver->setTextureGenFlags(video::TEXGEN_MIPMAPFILTER, video::FILTER_ANISOTROPIC);
    spDriver->setTextureGenFlags(video::TEXGEN_ANISOTROPY, 8);
    
    BrickHeadTex = spDriver->loadTexture("../textures/BrickHead.png");
    
    // Create the ground object
    GroundObj = spSmngr->createMesh(scene::MESH_PLANE);
    GroundObj->meshTransform(100.0f);
    GroundObj->addTexture(BrickHeadTex);
    GroundObj->textureTransform(0, 200.0f);
    GroundObj->setColorMaterial(false);
    GroundObj->setDiffuseColor(video::color(0, 150, 0));
    spCmngr->addPickingMesh(GroundObj);
}

void RenderGame()
{
    const io::stringc DefaultFilename = "../data/DefaultMap.bcm"; // Brick Coverage Map
    
    LoadMap(DefaultFilename);
    
    while (spDevice->updateEvent() && !spControl->keyDown(io::KEY_ESCAPE))
    {
        spDriver->clearBuffers();
        
        spSmngr->renderScene();
        
        UpdateGame();
        
        spDriver->flipBuffers();
    }
    
    SaveMap(DefaultFilename);
}

void UpdateGame()
{
    static bool CamMove;
    
    BasicBrick::UpdateBricks();
    
    if (spControl->mouseHit(io::MOUSE_MIDDLE))
        CamMove = !CamMove;
    if (spControl->mouseDown(io::MOUSE_RIGHT))
        CamMove = false;
    
    if (CamMove || spControl->mouseDown(io::MOUSE_RIGHT))
        tool::moveCameraFree(Cam, 0.25, 0.25, 90, false);
    
    if (CurDragedBrick && spControl->mouseHit(io::MOUSE_LEFT))
        CurDragedBrick = 0;
    
    if (spControl->keyHit(io::KEY_RETURN))
    {
        BasicBrick* NewBrick = BasicBrick::Create((EBasicBricks)CurBasicBrickType);
        
        if (NewBrick)
            NewBrick->SetColorType((EBrickColors)math::Rnd(COLOR_BROWN));
    }
    
    if (spControl->keyHit(io::KEY_PAGEUP))
    {
        if (++CurBasicBrickType > BRICK_4_3_2)
            CurBasicBrickType = BRICK_1_1_1;
        if (CurDragedBrick)
            CurDragedBrick->SetBasicBrickType((EBasicBricks)CurBasicBrickType);
    }
    if (spControl->keyHit(io::KEY_PAGEDOWN))
    {
        if (--CurBasicBrickType < BRICK_1_1_1)
            CurBasicBrickType = BRICK_4_3_2;
        if (CurDragedBrick)
            CurDragedBrick->SetBasicBrickType((EBasicBricks)CurBasicBrickType);
    }
    
    if (CurDragedBrick)
    {
        if (spControl->keyHit(io::KEY_LEFT))
            CurDragedBrick->TurnCCW();
        if (spControl->keyHit(io::KEY_RIGHT))
            CurDragedBrick->TurnCW();
    }
    
    if (spControl->keyHit(io::KEY_TAB))
    {
        static bool Frames;
        Frames = !Frames;
        spSmngr->setWireframe(Frames ? video::WIREFRAME_LINES : video::WIREFRAME_SOLID);
    }
    
    if (spControl->keyDown(io::KEY_CONTROL) && spControl->mouseHit(io::MOUSE_LEFT))
    {
        for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end(); ++it)
            (*it)->SetActive(false);
    }
    
    if (spControl->keyHit(io::KEY_DELETE))
    {
        if (CurDragedBrick)
        {
            for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end(); ++it)
            {
                if (*it == CurDragedBrick)
                {
                    spDeletePointer(CurDragedBrick);
                    BrickObjectList.erase(it);
                    break;
                }
            }
        }
        else
        {
            for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end();)
            {
                if ((*it)->GetActive())
                {
                    spDeletePointer(*it);
                    it = BrickObjectList.erase(it);
                }
                else
                    ++it;
            }
        }
    }
    
    if (spControl->keyDown(io::KEY_C))
    {
        EBrickColors ChoosenColor;
        
        if (BasicBrick::DrawMenuColorPalette(ChoosenColor))
        {
            for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end(); ++it)
            {
                if ((*it)->GetActive())
                    (*it)->SetColorType(ChoosenColor);
            }
        }
    }
    
    //BasicBrick::DrawMenuCurrentBrick();
}

void ClearGame()
{
    for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end(); ++it)
        spDeletePointer(*it);
    BrickObjectList.clear();
    
    deleteDevice();
}

static s8 BCMF_VERSION = 0x01;

bool SaveMap(const io::stringc &Filename)
{
    io::FileSystem FileSys;
    io::File* File = 0;
    
    if ( !( File = FileSys.openFile(Filename, io::FILE_WRITE) ) )
    {
        PrintError("Could not save file: \"" + Filename + "\"");
        return false;
    }
    
    File->writeString("BCMF");          // BrickCoverage Map File
    File->writeValue<s8>(BCMF_VERSION); // Format version
    
    File->writeValue<u32>(BrickObjectList.size());  // Count of brick objects
    
    for (std::list<BasicBrick*>::iterator it = BrickObjectList.begin(); it != BrickObjectList.end(); ++it)
    {
        File->writeValue<s8>((*it)->BasicType_);
        File->writeValue<s16>((*it)->Position_.X);
        File->writeValue<s16>((*it)->Position_.Y);
        File->writeValue<s16>((*it)->Position_.Z);
        
        switch ((*it)->Rotation_)
        {
            case   0: File->writeValue<s8>(0); break;
            case  90: File->writeValue<s8>(1); break;
            case 180: File->writeValue<s8>(2); break;
            case 270: File->writeValue<s8>(3); break;
        }
        
        File->writeValue<s8>((*it)->GetColorType());
        
        if ((*it)->GetColorType() == COLOR_EXTENDED)
        {
            File->writeValue<u8>((*it)->GetColor().Red);
            File->writeValue<u8>((*it)->GetColor().Green);
            File->writeValue<u8>((*it)->GetColor().Blue);
        }
    }
    
    FileSys.closeFile(File);
    
    return true;
}

bool LoadMap(const io::stringc &Filename)
{
    io::FileSystem FileSys;
    io::File* File = 0;
    
    if ( !( File = FileSys.openFile(Filename, io::FILE_READ) ) )
    {
        PrintError("Could not load file: \"" + Filename + "\"");
        return false;
    }
    
    if (File->readString(4) != "BCMF")
    {
        PrintError("Invalid magic number (must be \"BCMF\")");
        return false;
    }
    
    if (File->readValue<s8>() != BCMF_VERSION)
    {
        PrintError("Invalid file version (must be " + io::stringc((s32)BCMF_VERSION) + ")");
        return false;
    }
    
    const u32 ObjectCount = File->readValue<u32>();
    
    s8 BasicType;
    dim::vector3di Position;
    s8 Rotation;
    s8 ColorType;
    video::color Color;
    
    BasicBrick* NewBrick = 0;
    
    for (u32 i = 0; i < ObjectCount; ++i)
    {
        BasicType   = File->readValue<s8>();
        Position.X  = File->readValue<s16>();
        Position.Y  = File->readValue<s16>();
        Position.Z  = File->readValue<s16>();
        Rotation    = File->readValue<s8>();
        ColorType   = File->readValue<s8>();
        
        if (ColorType == COLOR_EXTENDED)
        {
            Color.Red   = File->readValue<u8>();
            Color.Green = File->readValue<u8>();
            Color.Blue  = File->readValue<u8>();
        }
        
        NewBrick = BasicBrick::Create((EBasicBricks)BasicType, true);
        
        if (NewBrick)
        {
            NewBrick->SetPosition(Position);
            
            switch (Rotation)
            {
                case 0: NewBrick->SetRotation(  0); break;
                case 1: NewBrick->SetRotation( 90); break;
                case 2: NewBrick->SetRotation(180); break;
                case 3: NewBrick->SetRotation(270); break;
            }
            
            if (ColorType == COLOR_EXTENDED)
                NewBrick->SetColor(Color);
            else
                NewBrick->SetColorType((EBrickColors)ColorType);
        }
    }
    
    FileSys.closeFile(File);
    
    return true;
}


} // /namespace bc








// ================================================================================
